from agton.jetton import JettonMaster

from ..types import SwapStep
from ..types import SwapStepParams
from ..types import SwapKind, GivenIn

class Pool(JettonMaster):
    def pack_swap_step(self,
                       limit: int,
                       next: SwapStep | None = None,
                       kind: SwapKind = GivenIn()) -> SwapStep:
        swap_step_params = SwapStepParams(kind, limit, next)
        return SwapStep(self.address, swap_step_params)



from __future__ import annotations

from dataclasses import dataclass
from typing import Self

from agton.ton import Builder, Slice, TlbConstructor

@dataclass(frozen=True, slots=True)
class GivenIn(TlbConstructor):
    @classmethod
    def tag(cls):
        return 0b0, 1

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        return cls()

    def serialize_fields(self, b: Builder) -> Builder:
        return b

@dataclass(frozen=True, slots=True)
class GivenOut(TlbConstructor):
    @classmethod
    def tag(cls):
        return 0b1, 1

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        return cls()

    def serialize_fields(self, b: Builder) -> Builder:
        return b

SwapKind = GivenIn | GivenOut

def swap_kind(s: Slice) -> SwapKind:
    b = s.preload_bit()
    match b:
        case 0: return GivenIn.deserialize(s)
        case 1: return GivenOut.deserialize(s)



from dataclasses import dataclass
from typing import Self

from agton.ton import Slice, Builder, TlbConstructor

from ..types import SwapStep, SwapParams

@dataclass(frozen=True, slots=True)
class SwapPayload(TlbConstructor):
    '''swap#e3a0d482 _:SwapStep swap_params:^SwapParams = ForwardPayload;'''
    swap_step: SwapStep
    swap_params: SwapParams

    @classmethod
    def tag(cls):
        return 0xe3a0d482, 32
    
    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        swap_step = s.load_tlb(SwapStep)
        swap_params = s.load_ref_tlb(SwapParams)
        return cls(swap_step, swap_params)
    
    def serialize_fields(self, b: Builder) -> Builder:
        return (
            b
            .store_tlb(self.swap_step)
            .store_ref_tlb(self.swap_params)
        )



from dataclasses import dataclass
from typing import Self

from agton.dedust.types import PoolParams
from agton.ton import Cell, Slice, Builder, TlbConstructor, begin_cell


@dataclass(frozen=True, slots=True)
class DepositLiquidity(TlbConstructor):
    '''
    deposit_liquidity#d55e4686 query_id:uint64 amount:Coins pool_params:PoolParams
                           ^[ min_lp_amount:Coins
                           asset0_target_balance:Coins asset1_target_balance:Coins ]
                           fulfill_payload:(Maybe ^Cell)
                           reject_payload:(Maybe ^Cell) = InMsgBody;
    '''
    query_id: int
    amount: int
    pool_params: PoolParams
    min_lp_amount: int
    asset0_target_balance: int
    asset1_target_balance: int
    fulfill_payload: Cell | None
    reject_payload: Cell | None

    @classmethod
    def tag(cls):
        return 0xd55e4686, 32

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        query_id = s.load_uint(64)
        amount = s.load_coins()
        pool_params = s.load_tlb(PoolParams)
        deposit = s.load_ref().begin_parse()
        min_lp_amount = deposit.load_coins()
        asset0_target_balance = deposit.load_coins()
        asset1_target_balance = deposit.load_coins()
        fulfill_payload = s.load_maybe_ref()
        reject_payload = s.load_maybe_ref()
        return cls(query_id, amount, pool_params, min_lp_amount,
                   asset0_target_balance, asset1_target_balance,
                   fulfill_payload, reject_payload)

    def serialize_fields(self, b: Builder) -> Builder:
        return (
            b
            .store_uint(self.query_id, 64)
            .store_coins(self.amount)
            .store_tlb(self.pool_params)
            .store_ref(
                begin_cell()
                .store_coins(self.min_lp_amount)
                .store_coins(self.asset0_target_balance)
                .store_coins(self.asset1_target_balance)
                .end_cell()
            )
            .store_maybe_ref(self.fulfill_payload)
            .store_maybe_ref(self.reject_payload)
        )



from .asset import Asset, Native, Jetton, ExtraCurrency
from .pool_params import PoolParams
from .pool_type import PoolType, Volatile, Stable
from .swap_kind import SwapKind, GivenIn, GivenOut
from .swap_params import SwapParams
from .swap_step import SwapStep, SwapStepParams



from dataclasses import dataclass
from typing import Self

from agton.dedust.types.swap_params import SwapParams
from agton.dedust.types.swap_step import SwapStep
from agton.ton.cell.builder import Builder
from agton.ton.cell.slice import Slice
from agton.ton.types.tlb import TlbConstructor


@dataclass(frozen=True, slots=True)
class Swap(TlbConstructor):
    '''
    swap#ea06185d query_id:uint64 amount:Coins _:SwapStep swap_params:^SwapParams = InMsgBody;
    '''
    query_id: int
    amount: int
    swap_step: SwapStep
    swap_params: SwapParams

    @classmethod
    def tag(cls):
        return 0xea06185d, 32

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        query_id = s.load_uint(64)
        amount = s.load_coins()
        swap_step = s.load_tlb(SwapStep.deserialize)
        swap_params = s.load_ref_tlb(SwapParams.deserialize)
        return cls(query_id, amount, swap_step, swap_params)

    def serialize_fields(self, b: Builder) -> Builder:
        b.store_uint(self.query_id, 64)
        b.store_coins(self.amount)
        b.store_tlb(self.swap_step)
        b.store_ref_tlb(self.swap_params)
        return b



from agton.ton import Cell, Contract, MessageRelaxed, CurrencyCollection, to_nano
from agton.ton.types import MsgAddress, MsgAddressInt, AddrNone
from agton.jetton import JettonWallet, JettonTransfer

from ..payloads import SwapPayload, DepositLiquidityPayload
from ..types import SwapStep
from ..types import SwapParams
from ..types import PoolParams

class JettonVault(Contract):
    def create_swap_params(self, 
                           recepient_addr: MsgAddressInt,
                           deadline: int = 0,
                           referral_addr: MsgAddress = AddrNone(),
                           fulfill_payload: Cell | None = None,
                           reject_payload: Cell | None = None) -> SwapParams:
        return SwapParams(
            deadline, recepient_addr, referral_addr, fulfill_payload, reject_payload
        )
    
    def create_swap_payload(self,
                            swap_step: SwapStep,
                            swap_params: SwapParams) -> SwapPayload:
        return SwapPayload(swap_step, swap_params)
    
    def create_swap_message(self,
                            jetton_wallet: JettonWallet,
                            query_id: int,
                            amount: int,
                            response_destination: MsgAddress,
                            swap_payload: SwapPayload,
                            value: int | CurrencyCollection = to_nano(0.5),
                            forward_amount: int = to_nano(0.3)) -> MessageRelaxed:
        return jetton_wallet.create_jetton_transfer(
            value=value,
            query_id=query_id,
            amount=amount,
            destination=self.address,
            response_destination=response_destination,
            forward_amount=forward_amount,
            forward_payload=swap_payload.to_cell()
        )
    
    def create_deposit_liquidity_payload(
        self,
        pool_params: PoolParams,
        min_lp_amount: int,
        asset0_target_balance: int,
        asset1_target_balance: int,
        fulfill_payload: Cell | None = None,
        reject_payload: Cell | None = None
    ) -> DepositLiquidityPayload:
        return DepositLiquidityPayload(
            pool_params, min_lp_amount,
            asset0_target_balance, asset1_target_balance,
            fulfill_payload, reject_payload
        )
    
    def create_deposit_liquidity_message(
        self,
        jetton_wallet: JettonWallet,
        query_id: int,
        amount: int,
        response_destination: MsgAddress,
        deposit_liquidity_payload: DepositLiquidityPayload,
        value: int | CurrencyCollection = to_nano(0.8),
        forward_amount: int = to_nano(0.5)
    ) -> MessageRelaxed:
        return jetton_wallet.create_jetton_transfer(
            value=value,
            query_id=query_id,
            amount=amount,
            destination=self.address,
            response_destination=response_destination,
            forward_amount=forward_amount,
            forward_payload=deposit_liquidity_payload.to_cell()
        )


from dataclasses import dataclass
from typing import Self

from agton.ton import Builder, Slice, TlbConstructor


@dataclass(frozen=True, slots=True)
class Volatile(TlbConstructor):
    ''' volatile$0 = PoolType; '''

    @classmethod
    def tag(cls) -> tuple[int, int] | None:
        return 0b0, 1

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        return cls()

    def serialize_fields(self, b: Builder) -> Builder:
        return b


@dataclass(frozen=True, slots=True)
class Stable(TlbConstructor):
    ''' stable$1 = PoolType; '''

    @classmethod
    def tag(cls) -> tuple[int, int] | None:
        return 0b0, 1

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        return cls()

    def serialize_fields(self, b: Builder) -> Builder:
        return b

PoolType = Volatile | Stable

def pool_type(s: Slice) -> PoolType:
    b = s.preload_bit()
    match b:
        case 0: return Volatile.deserialize(s)
        case 1: return Stable.deserialize(s)



from __future__ import annotations

from dataclasses import dataclass
from typing import Self

from agton.ton import TlbConstructor, MsgAddressInt, Address, Builder, Slice


@dataclass(frozen=True)
class Native(TlbConstructor):
    '''native$0000 = Asset;'''
    @classmethod
    def tag(cls) -> tuple[int, int] | None:
        return 0b0000, 4

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        return cls()

    def serialize_fields(self, b: Builder) -> Builder:
        return b


@dataclass(frozen=True, slots=True)
class Jetton(TlbConstructor):
    '''jetton$0001 workchain_id:int8 address:uint256 = Asset;'''
    address: MsgAddressInt

    @classmethod
    def tag(cls):
        return 0b0001, 4

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        wc = s.load_int(8)
        address = s.load_bytes(32)
        return cls(Address(wc, address))

    def serialize_fields(self, b: Builder) -> Builder:
        assert isinstance(self.address, Address)
        b.store_int(self.address.workchain, 8)
        b.store_bytes(self.address.address)
        return b


@dataclass(frozen=True)
class ExtraCurrency(TlbConstructor):
    '''extra_currency$0010 currency_id:int32 = Asset;'''
    currency_id: int

    @classmethod
    def tag(cls) -> tuple[int, int] | None:
        return 0b0010, 4

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        currency_id = s.load_int(32)
        return cls(currency_id)

    def serialize_fields(self, b: Builder) -> Builder:
        return b.store_int(self.currency_id, 32)


Asset = Native | Jetton | ExtraCurrency

def asset(s: Slice) -> Asset:
    tag = s.preload_uint(4)
    if tag == 0b0000:
        return Native.deserialize(s)
    elif tag == 0b0001:
        return Jetton.deserialize(s)
    elif tag == 0b0010:
        return ExtraCurrency.deserialize(s)
    else:
        raise ValueError(f"Unknown asset tag: {tag:#06x}")



from __future__ import annotations

from agton.ton import Contract, Provider, Address, MsgAddressInt, Slice, Cell

from ..types import SwapStep
from ..types import SwapStepParams
from ..types import SwapKind, GivenIn
from ..types import Asset, ExtraCurrency, Jetton, Native
from ..types import PoolType, Volatile

from .jetton_vault import JettonVault
from .native_vault import NativeVault
from .pool import Pool


class Factory(Contract):
    @classmethod
    def from_mainnet(cls, provider: Provider) -> Factory:
        a = Address.parse('EQBfBWT7X2BHg9tXAxzhz2aKiNTU1tpt5NsiK0uSDW_YAJ67')
        return cls(provider, a)

    def get_vault_address(self, asset: Asset) -> MsgAddressInt:
        s = self.run_get_method('get_vault_address', [asset.to_slice()])
        match s:
            case Slice():
                return s.load_msg_address_int()
            case Cell():
                return s.begin_parse().load_msg_address_int()
            case _:
                raise TypeError(f'Unexpected result for get_vault_address: {s!r}')

    def get_vault(self, asset: Asset) -> NativeVault | JettonVault:
        address = self.get_vault_address(asset)
        match asset:
            case Native(): 
                return NativeVault(self.provider, address)
            case Jetton():
                return JettonVault(self.provider, address)
            case ExtraCurrency():
                raise NotImplementedError('Not implemented by DeDust itself')
    
    def get_pool_address(self,
                         asset0: Asset,
                         asset1: Asset,
                         pool_type: PoolType | int = Volatile()) -> MsgAddressInt:
        if isinstance(pool_type, PoolType):
            pool_type = 0 if isinstance(pool_type, Volatile) else 1
        s = self.run_get_method('get_pool_address', [pool_type, asset0.to_slice(), asset1.to_slice()])

        match s:
            case Slice():
                return s.load_msg_address_int()
            case Cell():
                return s.begin_parse().load_msg_address_int()
            case _:
                raise TypeError(f'Unexpected result for get_pool_address: {s!r}')
    
    def get_pool(self,
                 asset0: Asset,
                 asset1: Asset,
                 pool_type: PoolType | int = Volatile()) -> Pool:
        a = self.get_pool_address(asset0, asset1, pool_type)
        return Pool(self.provider, a)



from dataclasses import dataclass
from typing import Self

from agton.ton import Builder, Slice, TlbConstructor

from .asset import Asset, asset
from .pool_type import PoolType, pool_type


@dataclass(frozen=True)
class PoolParams(TlbConstructor):
    '''
    pool_params#_ pool_type:PoolType asset0:Asset asset1:Asset = PoolParams;
    '''
    pool_type: PoolType
    asset0: Asset
    asset1: Asset

    @classmethod
    def tag(cls) -> tuple[int, int] | None:
        return None

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        t = s.load_tlb(pool_type)
        a0 = s.load_tlb(asset)
        a1 = s.load_tlb(asset)
        return cls(t, a0, a1)

    def serialize_fields(self, b: Builder) -> Builder:
        return (
            b
            .store_tlb(self.pool_type)
            .store_tlb(self.asset0)
            .store_tlb(self.asset1)
        )



from agton.ton import Contract

from agton.ton import Cell
from agton.ton.types import MessageRelaxed
from agton.ton.types import CurrencyCollection
from agton.ton.types import AddrNone, MsgAddress, MsgAddressInt

from ..messages import Swap, DepositLiquidity
from ..types import SwapStep
from ..types import SwapParams
from ..types import PoolParams


class NativeVault(Contract):
    def create_swap_params(self, 
                           recepient_addr: MsgAddressInt,
                           deadline: int = 0,
                           referral_addr: MsgAddress = AddrNone(),
                           fulfill_payload: Cell | None = None,
                           reject_payload: Cell | None = None) -> SwapParams:
        return SwapParams(
            deadline, recepient_addr, referral_addr, fulfill_payload, reject_payload
        )
    
    def create_swap_body(self, 
                         query_id: int,
                         amount: int,
                         swap_step: SwapStep,
                         swap_params: SwapParams) -> Swap:
         return Swap(query_id, amount, swap_step, swap_params)

    def create_swap_message(self,
                            value: int | CurrencyCollection,
                            swap_body: Swap) -> MessageRelaxed:
        swap_message = self.create_internal_message(
            value=value,
            body=swap_body.to_cell()
        )
        return swap_message
    
    def create_deposit_liquidity_body(self,
                                      query_id: int,
                                      amount: int,
                                      pool_params: PoolParams,
                                      min_lp_amount: int,
                                      asset0_target_balance: int,
                                      asset1_target_balance: int,
                                      fulfill_payload: Cell | None = None,
                                      reject_payload: Cell | None = None) -> DepositLiquidity:
        return DepositLiquidity(
            query_id, amount, pool_params,
            min_lp_amount, asset0_target_balance, asset1_target_balance,
            fulfill_payload, reject_payload
        )

    def create_deposit_liquidity_message(self,
                                         value: int | CurrencyCollection,
                                         deposit_liquidity_body: DepositLiquidity) -> MessageRelaxed:
            return self.create_internal_message(
                value=value,
                body=deposit_liquidity_body.to_cell(),
            )



from dataclasses import dataclass
from typing import Self

from agton.ton import Cell, Builder, Slice
from agton.ton import MsgAddressInt, MsgAddress
from agton.ton import TlbConstructor


@dataclass(frozen=True, slots=True)
class SwapParams(TlbConstructor):
    '''
    swap_params#_ deadline:Timestamp recipient_addr:MsgAddressInt referral_addr:MsgAddress
                  fulfill_payload:(Maybe ^Cell) reject_payload:(Maybe ^Cell) = SwapParams;
    '''
    deadline: int
    recipient_addr: MsgAddressInt
    referral_addr: MsgAddress
    fulfill_payload: Cell | None
    reject_payload: Cell | None

    @classmethod
    def tag(cls) -> tuple[int, int] | None:
        return None

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        deadline = s.load_uint(32)
        recipient_addr = s.load_msg_address_int()
        referral_addr = s.load_msg_address()
        fulfill_payload = s.load_maybe_ref()
        reject_payload = s.load_maybe_ref()
        return cls(deadline, recipient_addr, referral_addr, fulfill_payload, reject_payload)

    def serialize_fields(self, b: Builder) -> Builder:
        b.store_uint(self.deadline, 32)
        b.store_msg_address(self.recipient_addr)
        b.store_msg_address(self.referral_addr)
        b.store_maybe_ref(self.fulfill_payload)
        b.store_maybe_ref(self.reject_payload)
        return b




from __future__ import annotations

from dataclasses import dataclass
from typing import Self

from agton.ton import Slice, Builder, MsgAddressInt, MsgAddress
from agton.ton import TlbConstructor
from .types.asset import Asset


@dataclass(frozen=True, slots=True)
class Swap(TlbConstructor):
    '''
    swap#9c610de3 asset_in:Asset asset_out:Asset amount_in:Coins amount_out:Coins
                  ^[ sender_addr:MsgAddressInt referral_addr:MsgAddress
                  reserve0:Coins reserve1:Coins ] = ExtOutMsgBody;
    '''
    asset_in: Asset
    asset_out: Asset
    amount_in: int
    amount_out: int
    sender_addr: MsgAddressInt
    referral_addr: MsgAddress
    reserve0: int
    reserve1: int

    @classmethod
    def tag(cls):
        return 0x9c610de3, 32

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Swap:
        raise NotImplementedError

    def serialize_fields(self, b: Builder) -> Builder:
        raise NotImplementedError


@dataclass(frozen=True, slots=True)
class Deposit(TlbConstructor):
    '''
    deposit#b544f4a4 sender_addr:MsgAddressInt amount0:Coins amount1:Coins
                     reserve0:Coins reserve1:Coins liquidity:Coins = ExtOutMsgBody;
    '''
    sender_addr: MsgAddressInt
    amount0: int
    amount1: int
    reserve0: int
    reserve1: int
    liquidity: int

    @classmethod
    def tag(cls):
        return 0xb544f4a4, 32

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        raise NotImplementedError

    def serialize_fields(self, b: Builder) -> Builder:
        raise NotImplementedError


@dataclass(frozen=True)
class Withdrawal(TlbConstructor):
    '''
    withdrawal#3aa870a6 sender_addr:MsgAddressInt liquidity:Coins
                        amount0:Coins amount1:Coins
                        reserve0:Coins reserve1:Coins = ExtOutMsgBody;
    '''
    sender_addr: MsgAddressInt
    liquidity: int
    amount0: int
    amount1: int
    reserve0: int
    reserve1: int

    @classmethod
    def tag(cls):
        return 0x3aa870a6, 32

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        raise NotImplementedError

    def serialize_fields(self, b: Builder) -> Builder:
        raise NotImplementedError

Event = Swap | Deposit | Withdrawal

def event(s: Slice) -> Event:
    tag = s.preload_uint(32)
    if tag == Swap.tag()[0]:
        return Swap.deserialize(s)
    elif tag == Deposit.tag()[0]:
        return Deposit.deserialize(s)
    elif tag == Withdrawal.tag()[0]:
        return Withdrawal.deserialize(s)
    else:
        raise ValueError(f"Unknown dedust event tag: {tag:08x}")


from __future__ import annotations

from dataclasses import dataclass
from typing import Self, TYPE_CHECKING

from agton.ton import Builder, Slice, MsgAddressInt
from agton.ton import TlbConstructor

from .swap_kind import SwapKind, swap_kind


@dataclass(frozen=True, slots=True)
class SwapStepParams(TlbConstructor):
    '''
    step_params#_ kind:SwapKind limit:Coins next:(Maybe ^SwapStep) = SwapStepParams;
    '''
    kind: SwapKind
    limit: int
    next: SwapStep | None

    @classmethod
    def tag(cls) -> tuple[int, int] | None:
        return None

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        kind = s.load_tlb(swap_kind)
        limit = s.load_coins()
        next = s.load_maybe_ref_tlb(SwapStep.deserialize)
        return cls(kind, limit, next)

    def serialize_fields(self, b: Builder) -> Builder:
        b.store_tlb(self.kind)
        b.store_coins(self.limit)
        b.store_maybe_ref_tlb(self.next)
        return b


@dataclass(frozen=True, slots=True)
class SwapStep(TlbConstructor):
    '''
    step#_ pool_addr:MsgAddressInt params:SwapStepParams = SwapStep;
    '''
    pool_addr: MsgAddressInt
    params: SwapStepParams

    @classmethod
    def tag(cls) -> tuple[int, int] | None:
        return None

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        pool_addr = s.load_msg_address_int()
        params = s.load_tlb(SwapStepParams.deserialize)
        return cls(pool_addr, params)

    def serialize_fields(self, b: Builder) -> Builder:
        b.store_tlb(self.pool_addr)
        b.store_tlb(self.params)
        return b

